import unittest

from fakeredis import FakeStrictRedis

from cache_alchemy.config import DefaultConfig


class TestCacheConfig(DefaultConfig):
    ...


def get_config():
    config = TestCacheConfig()
    config.cache_redis_client = FakeStrictRedis.from_url(config.CACHE_ALCHEMY_REDIS_URL)
    return config


class CacheTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.config = get_config()
        self.config.cache_redis_client.flushdb()



from unittest.mock import MagicMock
from tests import get_config
from cache_alchemy import json_cache, memory_cache

redis_call_mock = MagicMock()

config = get_config()


@json_cache
def json_test():
    redis_call_mock()
    return "test"


memory_call_mock = MagicMock()


@memory_cache
def memory_test():
    memory_call_mock()
    return "test"



import time
import unittest
from typing import Type
from unittest.mock import Mock

from configalchemy.utils import import_reference

from cache_alchemy import (
    memory_cache,
    DefaultConfig,
    method_memory_cache,
    property_memory_cache,
)
from cache_alchemy.backends.memory import MemoryCache
from cache_alchemy.backends.memory import all_cache_pool
from cache_alchemy.lru import LRUDict
from cache_alchemy.utils import UnsupportedError
from tests import CacheTestCase


class MemoryCacheTestCase(CacheTestCase):
    def test_expire_and_limit(self):
        tmp = lambda: ...
        with self.assertRaises(TypeError):
            memory_cache(limit=-2)(tmp)
        with self.assertRaises(TypeError):
            memory_cache(expire=-2)(tmp)

        self.assertEqual(tmp, memory_cache(expire=0)(tmp))
        self.assertEqual(tmp, memory_cache(limit=0)(tmp))

        cached_tmp = memory_cache(limit=-1)(tmp)
        self.assertIsInstance(cached_tmp.cache.cache_pool, dict)
        self.assertNotIsInstance(cached_tmp.cache.cache_pool, LRUDict)
        with self.assertRaises(UnsupportedError):
            cached_tmp.cache_clear(1)

    def test_cache_class_method(self):
        call_mock = Mock()

        class Tmp:
            @classmethod
            @method_memory_cache
            def add(cls, a: int, b: int = 2) -> int:
                call_mock()
                return a + b

        self.assertEqual(Tmp.add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.add(2), 4)
        self.assertEqual(call_mock.call_count, 2)

    def test_cache_property(self):
        call_mock = Mock()
        name = "test"

        class Tmp:
            @property
            @property_memory_cache
            def name(self) -> str:
                call_mock()
                return name

        self.assertEqual(Tmp().name, name)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp().name, name)
        self.assertEqual(call_mock.call_count, 1)

    def test_distributed_memory_cache(self):
        call_mock = Mock()
        result = object()

        @memory_cache
        def add(a: int, b: int = 2) -> result:
            call_mock()
            return result

        self.assertEqual(0, add.cache_clear())
        self.assertEqual(result, add(1))
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(1), result)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(a=1), result)
        self.assertEqual(call_mock.call_count, 2)
        self.assertEqual(add(2), result)
        self.assertEqual(call_mock.call_count, 3)

        self.assertEqual(3, add.cache_clear())
        self.assertEqual(add(2), result)
        self.assertEqual(call_mock.call_count, 4)

    def test_distributed_strict_memory_cache(self):
        call_mock = Mock()
        result = object()

        @memory_cache(strict=True)
        def add(a: int, b: int = 2) -> result:
            call_mock()
            return result

        self.assertEqual(result, add(1))
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(1), result)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(a=1), result)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(2), result)
        self.assertEqual(call_mock.call_count, 2)

        add.cache_clear()
        self.assertEqual(add(1, b=2), result)
        self.assertEqual(call_mock.call_count, 3)

    def test_memory_cache(self):
        all_cache_pool.clear()

        class TestMemoryCacheConfig(DefaultConfig):
            CACHE_ALCHEMY_MEMORY_BACKEND = "cache_alchemy.backends.memory.MemoryCache"

        config = TestMemoryCacheConfig()
        call_mock = Mock()
        result = object()
        cache_backend: Type[MemoryCache] = import_reference(
            config.CACHE_ALCHEMY_MEMORY_BACKEND
        )

        self.assertEqual(set(), cache_backend.get_all_namespace())

        @memory_cache
        def add(a: int, b: int = 2) -> result:
            self.assertEqual(config, DefaultConfig.get_current_config())
            call_mock()
            return result

        self.assertIn(
            cache_backend.__name__, add.cache.make_key(args=(1,), kwargs={})[2]
        )
        self.assertEqual(0, len(cache_backend.get_all_namespace()))

        self.assertEqual(add(1), result)
        self.assertEqual(1, len(cache_backend.get_all_namespace()))
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(a=1), result)
        self.assertEqual(call_mock.call_count, 2)
        self.assertEqual(add(2), result)
        self.assertEqual(call_mock.call_count, 3)
        self.assertEqual(3, add.cache_clear())
        self.assertEqual(add(1), result)
        self.assertEqual(call_mock.call_count, 4)
        self.assertEqual(1, cache_backend.flush_cache())

    def test_strict_memory_cache(self):
        class TestMemoryCacheConfig(DefaultConfig):
            CACHE_ALCHEMY_MEMORY_BACKEND = "cache_alchemy.backends.memory.MemoryCache"

        config = TestMemoryCacheConfig()
        call_mock = Mock()
        result = object()

        @memory_cache(strict=True)
        def add(a: int, b: int = 2) -> result:
            self.assertEqual(config, DefaultConfig.get_current_config())
            call_mock()
            return result

        self.assertEqual(add(1), result)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(1), result)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(2), result)
        self.assertEqual(call_mock.call_count, 2)
        add.cache_clear()
        self.assertEqual(add(1), result)
        self.assertEqual(call_mock.call_count, 3)

    def test_cache_namespace_hash(self):
        from tests import resource_a, resource_b

        resource_a.memory_test()
        resource_b.memory_test()
        resource_a.memory_test()
        resource_b.memory_test()
        self.assertEqual(1, resource_a.memory_call_mock.call_count)
        self.assertEqual(1, resource_b.memory_call_mock.call_count)

    def test_cache_clear_with_pattern(self):
        class TestMemoryCacheConfig(DefaultConfig):
            CACHE_ALCHEMY_MEMORY_BACKEND = "cache_alchemy.backends.memory.MemoryCache"

        config = TestMemoryCacheConfig()
        call_mock = Mock()
        result = object()

        @memory_cache(strict=True)
        def add(a: int, b: int = 2) -> result:
            self.assertEqual(config, DefaultConfig.get_current_config())
            call_mock()
            return result

        add(1)
        self.assertEqual(1, call_mock.call_count)
        add(a=2)
        self.assertEqual(2, call_mock.call_count)
        self.assertEqual(1, add.cache_clear(a=1))
        add(2)
        self.assertEqual(2, call_mock.call_count)
        add(1)
        self.assertEqual(3, call_mock.call_count)

    def test_distributed_strict_memory_cache_clear_with_pattern(self):
        call_mock = Mock()
        result = object()

        @memory_cache(strict=True)
        def add(a: int, b: int = 2) -> result:
            call_mock()
            return result

        add(1)
        self.assertEqual(1, call_mock.call_count)
        add(a=2)
        self.assertEqual(2, call_mock.call_count)
        add.cache_clear(a=1)
        add(2)
        self.assertEqual(2, call_mock.call_count)
        add(1)
        self.assertEqual(3, call_mock.call_count)

    def test_memory_cache_expire(self):
        call_mock = Mock()

        @memory_cache(expire=1)
        def add(a: int, b: int = 2) -> int:
            call_mock()
            return a + b

        self.assertEqual(add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        time.sleep(3)
        self.assertEqual(add(1), 3)
        self.assertEqual(call_mock.call_count, 2)


if __name__ == "__main__":
    unittest.main()



"""
Caching framework.
"""
__version__ = "0.4.3"

from functools import wraps
from importlib import import_module
from types import FunctionType
from typing import Callable, List, Optional, cast, Type, TypeVar

from .backends.base import BaseCache, CacheFunctionType
from .config import DefaultConfig
from .dependency import CacheDependency
from .lru import LRUDict
from .utils import UnsupportedError

BackendCls = TypeVar("BackendCls", bound=BaseCache)


def _create_cache(
    *,
    backend_cls: Type[BackendCls],
    cached_function: FunctionType,
    expire: Optional[int],
    limit: Optional[int],
    is_method: bool = False,
    strict=False,
    **kwargs,
) -> BackendCls:

    if not isinstance(limit, int) or limit < -1:
        raise TypeError("Expected limit to larger equal than -1")
    if not isinstance(expire, int) or expire < -1:
        raise TypeError("Expected expire to larger equal than -1")

    return backend_cls(  # type: ignore
        cached_function=cached_function,
        expire=expire,
        limit=limit,
        is_method=is_method,
        strict=strict,
        **kwargs,
    )


CacheDecoratorType = Callable[[FunctionType], CacheFunctionType]


def cache(
    limit: Optional[int],
    expire: Optional[int],
    is_method: bool,
    strict: bool,
    backend: str,
    dependency: List[CacheDependency],
    **kwargs,
) -> CacheDecoratorType:
    """The base function to creat a cache object like this::

        @cache(
            limit=1000,
            expire=60,
            is_method=False,
            strict=True,
            backend="cache_alchemy.backends.memory.MemoryCache",
            dependency=[],
        )
        def f(x, y):
            pass

        # To clear cache
        f.cache_clear()

    :param int expire: expire time with an integer value used as seconds.
    :param bool is_method: If *True*, the first argument will be ignored in generate cache key.
    :param bool strict: If *False*, make a cache key in a way that is flat as possible rather than
                        as a nested and strict structure that would support partially cache clear.
                        it means that f(x=1, y=2) will now be treated as a distinct call from
                        f(y=2, x=1) which will be cached separately.
    """
    config = DefaultConfig.get_current_config()
    module_path, class_name = backend.rsplit(".", 1)
    backend_cls = getattr(import_module(module_path), class_name)

    if limit is None:
        limit = config.CACHE_ALCHEMY_DEFAULT_LIMIT
    if expire is None:
        expire = config.CACHE_ALCHEMY_DEFAULT_EXPIRE

    def decorating_function(func: CacheFunctionType) -> CacheFunctionType:
        if limit == 0 or expire == 0:
            return func

        cache = _create_cache(
            backend_cls=backend_cls,
            expire=expire,
            cached_function=cast(FunctionType, func),
            limit=limit,
            is_method=is_method,
            strict=strict,
            **kwargs,
        )

        @wraps(func)
        def wrapper(*args, **kwargs):
            return cache(*args, **kwargs)

        def cache_clear(*args, **kwargs) -> int:
            """Clear the cache and cache statistics"""
            if not strict and (args or kwargs):
                raise UnsupportedError("fast hash not support pattern delete")

            return sum(
                [
                    cache.cache_clear(args, kwargs),
                    CacheDependency.dependent_cache_clear(cache, args, kwargs),
                ]
            )

        for item in dependency:
            item.cache_objects.add(cache)
            item.register_dependency(item)

        wrapper.cache = cache  # type: ignore
        wrapper.cache_clear = cache_clear  # type: ignore
        return wrapper

    if callable(limit):
        user_function, limit = limit, config.CACHE_ALCHEMY_DEFAULT_LIMIT
        return decorating_function(user_function)
    return decorating_function


def json_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    is_method: bool = False,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=is_method,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_JSON_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def method_json_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=True,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_JSON_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def property_json_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=True,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_JSON_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def memory_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    is_method: bool = False,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=is_method,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_MEMORY_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def method_memory_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=True,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_MEMORY_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def property_memory_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=True,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_MEMORY_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def pickle_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    is_method: bool = False,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=is_method,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_PICKLE_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def method_pickle_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=True,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_PICKLE_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )


def property_pickle_cache(
    limit: Optional[int] = None,
    *,
    expire: Optional[int] = None,
    strict: bool = False,
    dependency: Optional[List[CacheDependency]] = None,
    **kwargs,
) -> CacheDecoratorType:
    return cache(
        limit=limit,
        expire=expire,
        is_method=True,
        strict=strict,
        backend=DefaultConfig.get_current_config().CACHE_ALCHEMY_PICKLE_BACKEND,
        dependency=dependency or [],
        **kwargs,
    )



import unittest

from cache_alchemy import DefaultConfig


class ConfigTestCase(unittest.TestCase):
    def test_get_current_config(self):
        with self.assertRaises(RuntimeError):
            DefaultConfig.get_current_config()

        class CacheConfig(DefaultConfig):
            ...

        cache_config = CacheConfig()
        self.assertEqual(cache_config, DefaultConfig.get_current_config())


if __name__ == "__main__":
    unittest.main()



import time
import unittest
from typing import Type
from unittest.mock import Mock

from configalchemy.utils import import_reference

from cache_alchemy import json_cache, method_json_cache, property_json_cache
from cache_alchemy.backends.json import DistributedJsonCache
from cache_alchemy.utils import UnsupportedError
from tests import CacheTestCase


class JsonCacheTestCase(CacheTestCase):
    def test_cache_function(self):
        call_mock = Mock()

        @json_cache
        def add(a: int, b: int = 2) -> int:
            call_mock()
            return a + b

        self.assertEqual(0, add.cache_clear())
        self.assertEqual(add(1), 3)
        self.assertEqual(1, call_mock.call_count)
        self.assertEqual(1, add.cache.misses)
        self.assertEqual(0, add.cache.hits)
        self.assertEqual(add(a=1), 3)
        self.assertEqual(2, add.cache.misses)
        self.assertEqual(0, add.cache.hits)
        self.assertEqual(2, call_mock.call_count)
        self.assertEqual(add(a=2), 4)
        self.assertEqual(3, call_mock.call_count)
        self.assertEqual(3, add.cache.misses)
        self.assertEqual(0, add.cache.hits)
        self.assertEqual(3, add.cache_clear())
        self.assertEqual(4, add(a=2))
        self.assertEqual(4, call_mock.call_count)

    def test_cache_namespace_hash(self):
        from tests import resource_a, resource_b

        resource_a.json_test()
        resource_b.json_test()
        resource_a.json_test()
        resource_b.json_test()
        self.assertEqual(1, resource_a.redis_call_mock.call_count)
        self.assertEqual(1, resource_b.redis_call_mock.call_count)

    def test_cache_clear(self):
        call_mock = Mock()

        @json_cache
        def add(a: int, b: int = 2) -> int:
            call_mock()
            return a + b

        self.assertEqual(add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        add.cache_clear()
        self.assertEqual(add(a=1), 3)
        self.assertEqual(call_mock.call_count, 2)

        with self.assertRaises(UnsupportedError):
            add.cache_clear(a=1)

    def test_cache_set(self):
        self.config.cache_redis_client.flushdb()
        json_cache_backend: Type[DistributedJsonCache] = import_reference(
            self.config.CACHE_ALCHEMY_JSON_BACKEND
        )

        self.assertEqual(set(), json_cache_backend.get_all_namespace())

        @json_cache
        def add(a: int, b: int = 2) -> int:
            return a + b

        @json_cache
        def mul(a: int, b: int = 2) -> int:
            return a * b

        self.assertIn(
            json_cache_backend.__name__, add.cache.make_key(args=(1,), kwargs={})[2]
        )
        self.assertEqual(0, len(json_cache_backend.get_all_namespace()))
        self.assertEqual(0, json_cache_backend.flush_cache())
        add(1)
        self.assertEqual(1, len(json_cache_backend.get_all_namespace()))
        self.assertEqual(1, json_cache_backend.flush_cache())
        add(1)
        self.assertEqual(2, add.cache.misses)
        mul(1)
        self.assertEqual(2, len(json_cache_backend.get_all_namespace()))
        self.assertEqual(2, json_cache_backend.flush_cache())

    def test_cache_clear_with_pattern_and_int_parameter(self):
        call_mock = Mock()

        @json_cache(strict=True)
        def add(a: int, b: int = 2) -> int:
            call_mock()
            return a + b

        add(1)
        self.assertEqual(1, call_mock.call_count)
        add(a=2)
        self.assertEqual(2, call_mock.call_count)
        self.assertEqual(1, add.cache_clear(a=1))
        add(2)
        self.assertEqual(2, call_mock.call_count)
        add(1)
        self.assertEqual(3, call_mock.call_count)
        self.assertEqual(0, add.cache_clear(a=3))

    def test_cache_clear_with_pattern_and_str_parameter(self):
        call_mock = Mock()

        @json_cache(strict=True)
        def hello(name: str, status: str = "") -> str:
            call_mock()
            return ""

        hello(name="world", status="test")
        hello(name="world", status="prod")
        self.assertEqual(2, call_mock.call_count)
        self.assertEqual(1, hello.cache_clear(name="world", status="prod"))

    def test_strict_cache_function(self):
        call_mock = Mock()

        @json_cache(strict=True)
        def add(a: int, b: int = 2) -> int:
            call_mock()
            return a + b

        self.assertEqual(add(1), 3)
        self.assertEqual(1, call_mock.call_count)
        self.assertEqual(1, add.cache.misses)
        self.assertEqual(0, add.cache.hits)
        self.assertEqual(add(a=1), 3)
        self.assertEqual(1, add.cache.misses)
        self.assertEqual(1, add.cache.hits)
        self.assertEqual(1, call_mock.call_count)
        self.assertEqual(add(a=2), 4)
        self.assertEqual(2, call_mock.call_count)
        self.assertEqual(add.cache.misses, 2)
        self.assertEqual(add.cache.hits, 1)
        add.cache_clear()
        self.assertEqual(add(a=2), 4)
        self.assertEqual(3, call_mock.call_count)

    def test_cache_method(self):
        call_mock = Mock()

        class Tmp:
            @method_json_cache
            def add(self, a: int, b: int = 2) -> int:
                call_mock()
                return a + b

        self.assertEqual(Tmp().add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp().add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp().add(2), 4)
        self.assertEqual(call_mock.call_count, 2)

    def test_cache_class_method(self):
        call_mock = Mock()

        class Tmp:
            @classmethod
            @method_json_cache
            def add(cls, a: int, b: int = 2) -> int:
                call_mock()
                return a + b

        self.assertEqual(Tmp.add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.add(2), 4)
        self.assertEqual(call_mock.call_count, 2)

    def test_cache_static_method(self):
        call_mock = Mock()

        class Tmp:
            @staticmethod
            @json_cache
            def add(a: int, b: int = 2) -> int:
                call_mock()
                return a + b

        self.assertEqual(Tmp.add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.add(2), 4)
        self.assertEqual(call_mock.call_count, 2)

    def test_cache_property(self):
        call_mock = Mock()
        name = "test"

        class Tmp:
            @property
            @property_json_cache
            def name(self) -> str:
                call_mock()
                return name

        self.assertEqual(Tmp().name, name)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp().name, name)
        self.assertEqual(call_mock.call_count, 1)

    def test_json_cache_limit(self):
        call_mock = Mock()

        @json_cache(limit=1)
        def add(a: int, b: int = 2) -> int:
            call_mock()
            return a + b

        self.assertEqual(add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(add(a=2), 4)
        self.assertEqual(call_mock.call_count, 2)
        self.assertEqual(add(1), 3)
        self.assertEqual(call_mock.call_count, 3)

    def test_json_cache_expire(self):
        call_mock = Mock()

        @json_cache(expire=1)
        def add(a: int, b: int = 2) -> int:
            call_mock()
            return a + b

        self.assertEqual(add(1), 3)
        self.assertEqual(call_mock.call_count, 1)
        time.sleep(3)
        self.assertEqual(add(1), 3)
        self.assertEqual(call_mock.call_count, 2)

        unexpired_add_call_mock = Mock()

        @json_cache(expire=-1)
        def unexpired_add(a: int, b: int = 2) -> int:
            unexpired_add_call_mock()
            return a + b

        self.assertEqual(unexpired_add(1), 3)
        self.assertEqual(unexpired_add_call_mock.call_count, 1)
        self.assertEqual(unexpired_add(1), 3)
        self.assertEqual(unexpired_add_call_mock.call_count, 1)


if __name__ == "__main__":
    unittest.main()



import re
import unittest
from typing import List, Tuple

from cache_alchemy.utils import (
    generate_strict_key,
    generate_fast_key,
    generate_strict_key_pattern,
    UnsupportedError,
    generate_fast_key_pattern,
    escape,
)


class UtilsTestCase(unittest.TestCase):
    def test_generate_strict_key(self):
        class Tmp:
            @classmethod
            def tmp_class_method(cls, a, b=1):
                ...

            def tmp_method(self, a, b=1):
                ...

        tmp = Tmp()
        test_data: List[Tuple[str, dict]] = [
            ("", dict(args=(), kwargs=dict(), func=lambda *a: ...)),
            ("a01", dict(args=(1,), kwargs=dict(), func=lambda *a: ...)),
            ("a01a12", dict(args=(1, 2), kwargs=dict(), func=lambda *a: ...)),
            ("a'1'b2", dict(args=(), kwargs=dict(a="1", b=2), func=lambda **k: ...)),
            ("a'1'", dict(args=(), kwargs=dict(a="1"), func=lambda **k: ...)),
            ("", dict(args=(), kwargs=dict(), func=lambda **k: ...)),
            ("a01a12b2", dict(args=(1, 2), kwargs=dict(b=2), func=lambda *a, **k: ...)),
            ("a1b1", dict(args=(1,), kwargs=dict(), func=lambda a, b=1: ...)),
            ("a1b1", dict(args=(), kwargs=dict(a=1), func=lambda a, b=1: ...)),
            ("a1b2", dict(args=(1,), kwargs=dict(b=2), func=lambda a, b=1: ...)),
            ("a'1'b'2'", dict(args=("1",), kwargs=dict(b="2"), func=lambda a, b: ...)),
            (
                "a'1'b'2'",
                dict(args=("1",), kwargs=dict(b="2"), func=lambda a, b="": ...),
            ),
            ("a1b2", dict(args=(1,), kwargs=dict(b=2), func=lambda a, *, b=1: ...)),
            ("a1b1", dict(args=(), kwargs=dict(), func=lambda a=1, b=1: ...)),
            ("a3b1", dict(args=(3,), kwargs=dict(), func=lambda a=1, b=1: ...)),
            ("a3b2", dict(args=(3,), kwargs=dict(b=2), func=lambda a=1, b=1: ...)),
            ("a1b2", dict(args=(), kwargs=dict(b=2), func=lambda a=1, b=1: ...)),
            (
                f"cls{Tmp}a1b1",
                dict(args=(Tmp, 1), kwargs=dict(), func=tmp.tmp_class_method.__func__),
            ),
            (
                f"cls{Tmp}a1b2",
                dict(
                    args=(Tmp, 1), kwargs=dict(b=2), func=tmp.tmp_class_method.__func__
                ),
            ),
            (
                "a1b1",
                dict(
                    args=(tmp, 1),
                    kwargs=dict(),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
            ),
            (
                "a1b3",
                dict(
                    args=(tmp, 1, 3),
                    kwargs=dict(),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
            ),
            (
                "a1b2",
                dict(
                    args=(tmp, 1),
                    kwargs=dict(b=2),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
            ),
        ]
        for excepted_key, data in test_data:
            with self.subTest(excepted_key=excepted_key, data=data):
                self.assertEqual(excepted_key, generate_strict_key(**data)[2])

    def test_generate_strict_key_pattern(self):
        class Tmp:
            @classmethod
            def tmp_class_method(cls, a, b=1):
                ...

            def tmp_method(self, a, b=1):
                ...

        tmp = Tmp()
        test_data: List[Tuple[str, dict, List[str], List[str]]] = [
            (".*?", dict(args=(), kwargs=dict(), func=lambda *a: ...), ["a02a13"], []),
            (
                "a01.*?",
                dict(args=(1,), kwargs=dict(), func=lambda *a: ...),
                ["a01a11"],
                ["a02a11"],
            ),
            (
                "a'1'b2.*?",
                dict(args=(), kwargs=dict(a="1", b=2), func=lambda **k: ...),
                ["a'1'b2c3"],
                ["a'2'b2c3"],
            ),
            (
                "a'1'.*?",
                dict(args=(), kwargs=dict(a="1"), func=lambda **k: ...),
                ["a'1'b2c3"],
                ["a2b2"],
            ),
            (
                "a01a12.*?b2.*?",
                dict(args=(1, 2), kwargs=dict(b=2), func=lambda *a, **k: ...),
                ["a01a12a23b2c3"],
                ["a02"],
            ),
            (
                "a1b.*?",
                dict(args=(1,), kwargs=dict(), func=lambda a, b: ...),
                ["a1b2"],
                ["a2b2"],
            ),
            (
                "a.*?b1",
                dict(args=(), kwargs=dict(b=1), func=lambda a, b: ...),
                ["a1b1"],
                ["a1b2"],
            ),
            (
                "a1b.*?",
                dict(args=(), kwargs=dict(a=1), func=lambda a, b: ...),
                ["a1b1"],
                ["a2b1"],
            ),
            (
                "a1b2",
                dict(args=(1,), kwargs=dict(b=2), func=lambda a, b: ...),
                ["a1b2"],
                ["a2b2", "a1b3"],
            ),
            (
                "a1b.*?",
                dict(args=(1,), kwargs=dict(), func=lambda a, b=1: ...),
                ["a1b1"],
                ["a2b2"],
            ),
            (
                "a1b1",
                dict(args=(1,), kwargs=dict(b=1), func=lambda a, b=1: ...),
                ["a1b1"],
                ["a2b2"],
            ),
            (
                "a1b2",
                dict(args=(1,), kwargs=dict(b=2), func=lambda a, *, b=1: ...),
                ["a1b2"],
                ["a2b2"],
            ),
            (
                "a1b.*?",
                dict(args=(1,), kwargs=dict(), func=lambda a, *, b=1: ...),
                ["a1b1"],
                ["a2b2"],
            ),
            (
                "a.*?b2",
                dict(args=(), kwargs=dict(b=2), func=lambda a=1, b=1: ...),
                ["a1b2"],
                ["a1b1"],
            ),
            (
                "a1b.*?",
                dict(args=(1,), kwargs=dict(), func=lambda a=1, b=1: ...),
                ["a1b1"],
                ["a2b2"],
            ),
            (
                f"cls{escape(str(Tmp))}a1b.*?",
                dict(args=(Tmp, 1), kwargs=dict(), func=tmp.tmp_class_method.__func__),
                [f"cls{Tmp}a1b1"],
                [f"cls{Tmp}a2b1"],
            ),
            (
                f"cls{escape(str(Tmp))}a1b2",
                dict(
                    args=(Tmp, 1), kwargs=dict(b=2), func=tmp.tmp_class_method.__func__
                ),
                [f"cls{Tmp}a1b2"],
                [f"cls{Tmp}a1b3"],
            ),
            (
                "a1b.*?",
                dict(
                    args=(tmp, 1),
                    kwargs=dict(),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
                ["a1b1"],
                ["a2b2"],
            ),
            (
                "a1b3",
                dict(
                    args=(tmp, 1, 3),
                    kwargs=dict(),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
                ["a1b3"],
                ["a2b3"],
            ),
            (
                "a1b2",
                dict(
                    args=(tmp, 1),
                    kwargs=dict(b=2),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
                ["a1b2"],
                ["a2b3"],
            ),
        ]
        for excepted_pattern, data, hit_keys, miss_keys in test_data:
            with self.subTest(excepted_pattern=excepted_pattern, data=data):
                excepted_pattern += "$"
                self.assertEqual(
                    excepted_pattern, generate_strict_key_pattern(**data),
                )
                pattern = re.compile(excepted_pattern, re.DOTALL)
                for key in hit_keys:
                    self.assertTrue(bool(pattern.match(key)))
                for key in miss_keys:
                    self.assertFalse(bool(pattern.match(key)))

    def test_generate_fast_key(self):
        class Tmp:
            @classmethod
            def tmp_class_method(cls, a, b=1):
                ...

            def tmp_method(self, a, b=1):
                ...

        tmp = Tmp()
        test_data: List[Tuple[str, dict]] = [
            ("()", dict(args=(), kwargs=dict(), func=lambda *a: ...)),
            ("(1,)", dict(args=(1,), kwargs=dict(), func=lambda *a: ...)),
            ("(1, 2)", dict(args=(1, 2), kwargs=dict(), func=lambda *a: ...)),
            (
                "('a', '1', 'b', 2)",
                dict(args=(), kwargs=dict(a="1", b=2), func=lambda **k: ...),
            ),
            ("('a', '1')", dict(args=(), kwargs=dict(a="1"), func=lambda **k: ...)),
            ("()", dict(args=(), kwargs=dict(), func=lambda **k: ...)),
            (
                "(1, 2, 'b', 2)",
                dict(args=(1, 2), kwargs=dict(b=2), func=lambda *a, **k: ...),
            ),
            ("(1,)", dict(args=(1,), kwargs=dict(), func=lambda a, b=1: ...)),
            ("('a', 1)", dict(args=(), kwargs=dict(a=1), func=lambda a, b=1: ...)),
            ("(1, 'b', 2)", dict(args=(1,), kwargs=dict(b=2), func=lambda a, b=1: ...)),
            (
                "(1, 'b', 2)",
                dict(args=(1,), kwargs=dict(b=2), func=lambda a, *, b=1: ...),
            ),
            (
                "(1, 'b', 2)",
                dict(args=(1,), kwargs=dict(b=2), func=lambda a=1, b=1: ...),
            ),
            ("('b', 2)", dict(args=(), kwargs=dict(b=2), func=lambda a=1, b=1: ...)),
            (
                f"({Tmp}, 1)",
                dict(args=(Tmp, 1), kwargs=dict(), func=tmp.tmp_class_method.__func__),
            ),
            (
                f"({Tmp}, 1, 'b', 2)",
                dict(
                    args=(Tmp, 1), kwargs=dict(b=2), func=tmp.tmp_class_method.__func__
                ),
            ),
            (
                "(1,)",
                dict(
                    args=(tmp, 1),
                    kwargs=dict(),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
            ),
            (
                "(1, 'b', 2)",
                dict(
                    args=(tmp, 1),
                    kwargs=dict(b=2),
                    func=tmp.tmp_method.__func__,
                    is_method=True,
                ),
            ),
        ]
        for excepted_key, data in test_data:
            with self.subTest(excepted_key=excepted_key, data=data):
                self.assertEqual(excepted_key, generate_fast_key(**data)[2])

    def test_generate_fast_key_pattern(self):
        with self.assertRaises(UnsupportedError):
            generate_fast_key_pattern(
                args=(), kwargs=dict(), func=lambda: None, is_method=True
            )


if __name__ == "__main__":
    unittest.main()



import pickle

from .base import DistributedCache, BaseCache, ReturnType


class DistributedPickleCache(DistributedCache, BaseCache[ReturnType]):
    def serialize(self, value: ReturnType) -> bytes:
        return pickle.dumps(value)

    def deserialize(self, result: bytes) -> ReturnType:
        return pickle.loads(result)



import unittest
from unittest.mock import Mock

from tests import CacheTestCase
from cache_alchemy import json_cache
from cache_alchemy.dependency import FunctionCacheDependency


class DependencyTestCase(CacheTestCase):
    def test_function_dependency(self):
        call_mock = Mock()

        @json_cache
        def add(a, b):
            return a + b

        dependency = FunctionCacheDependency(add)

        @json_cache(dependency=[dependency])
        def add_and_double(a, b):
            call_mock()
            return add(a, b) * 2

        self.assertEqual(4, add_and_double(1, 1))
        self.assertEqual(1, call_mock.call_count)
        add.cache_clear()
        self.assertEqual(4, add_and_double(1, 1))
        self.assertEqual(2, call_mock.call_count)


if __name__ == "__main__":
    unittest.main()



from threading import Lock
from typing import Iterable

_sentry = object()


class DoubleLinkedList(Iterable):
    __slots__ = ("prev", "next", "key", "result")

    def __init__(
        self, key, result=None,
    ):
        self.prev = self
        self.next = self
        self.key = key
        self.result = result

    def __iter__(self):
        if self.prev.key is _sentry:
            return
        yield self.prev
        yield from self.prev

    def __hash__(self):
        return hash(self.key)

    def __len__(self):
        return sum(map(lambda _: 1, self))

    def __repr__(self):
        return repr(self.result)

    def __str__(self):
        return str(self.result)


class LRUDict(dict):
    __slots__ = ("max_size", "root", "lock", "__weakref__")

    def __init__(self, max_size: int):
        if max_size <= 0:
            raise TypeError("Expected max_size to be larger than 0")
        self.max_size = max_size
        self.root = DoubleLinkedList(key=_sentry)
        self.lock = Lock()
        super().__init__()

    @property
    def full(self) -> bool:
        return len(self) >= self.max_size

    def __setitem__(self, key, value):
        with self.lock:
            if self.full:
                # Use the old root to store the new key and result.
                oldroot = self.root
                oldroot.key = key
                oldroot.result = value
                # Empty the oldest link and make it the new root.
                # Keep a reference to the old key and old result to
                # prevent their ref counts from going to zero during the
                # update. That will prevent potentially arbitrary object
                # clean-up code (i.e. __del__) from running while we're
                # still adjusting the links.
                self.root = oldroot.next
                oldkey = self.root.key
                self.root.key = self.root.result = _sentry
                # Now update the cache dictionary.
                del self[oldkey]
                # Save the potentially reentrant cache[key] assignment
                # for last, after the root and links have been put in
                # a consistent state.
                super().__setitem__(key, oldroot)
            else:
                last = self.root.prev
                link = DoubleLinkedList(key=key, result=value)
                link.prev = last
                link.next = self.root
                last.next = self.root.prev = link
                super().__setitem__(key, link)

    def __getitem__(self, item):
        with self.lock:
            value = super().__getitem__(item)
            value.prev.next = value.next
            value.next.prev = value.prev
            last = self.root.prev
            last.next = self.root.prev = value
            value.prev = last
            value.next = self.root
            return value.result

    def get(self, k, default=None):
        """Use EAFP to avoid RLock"""
        try:
            return self[k]
        except KeyError:
            return default

    def clear(self):
        with self.lock:
            self.root = DoubleLinkedList(key=_sentry)
            super().clear()



from timeit import timeit

from cache_alchemy.utils import generate_fast_key, generate_strict_key

__all__ = ["generate_fast_key", "generate_strict_key"]
tmp = lambda a, b=1, *, c, d=2, **kwd: ...


def benchmark():
    print(
        timeit(
            "fast_generate_key(args=(1,2), kwargs=dict(c=1,d=2), func=tmp)",
            globals=globals(),
        )
    )
    print(
        timeit(
            "strict_generate_key(args=(), kwargs=dict(a=1,b=2, c=1,d=2,e=3), func=tmp)",
            globals=globals(),
        )
    )


if __name__ == "__main__":
    benchmark()



import unittest

from cache_alchemy import LRUDict
from cache_alchemy.lru import DoubleLinkedList, _sentry


class LRUTestCase(unittest.TestCase):
    def test_dict_full(self):
        with self.assertRaises(TypeError):
            LRUDict(-1)
        lru_dict = LRUDict(1)
        lru_dict[1] = 1
        self.assertTrue(lru_dict.full)
        self.assertEqual(1, lru_dict[1])
        lru_dict[2] = 2
        self.assertFalse(1 in lru_dict)
        self.assertEqual(2, lru_dict[2])
        self.assertEqual("{2: 2}", str(lru_dict))
        self.assertEqual("{2: 2}", repr(lru_dict))

    def test_lru_dict(self):
        lru_dict = LRUDict(5)
        for index in range(5):
            lru_dict[index] = index
        for index, link in zip(range(4, -1, -1), lru_dict.root):
            self.assertEqual(index, link.key)
        for index in range(4, -1, -1):
            self.assertEqual(index, lru_dict[index])
        for index, link in zip(range(5), lru_dict.root):
            self.assertEqual(index, link.key)

    def test_lru_dict_clear(self):
        lru_dict = LRUDict(5)
        for index in range(5):
            lru_dict[index] = index
        lru_dict.clear()
        self.assertEqual(0, len(lru_dict.root))

    def test_double_link(self):
        root = DoubleLinkedList(key=_sentry)
        last = root.prev
        link = DoubleLinkedList(key="1", result=1)
        link.prev = last
        link.next = root
        root.prev = last.next = link
        self.assertEqual([link], list(root))
        self.assertEqual(1, len(root))
        self.assertEqual("1", str(link))


if __name__ == "__main__":
    unittest.main()



import time
from typing import Any, Callable, Dict, TypeVar, Set, Optional

from .base import BaseCache, DistributedCache
from ..lru import LRUDict

ReturnType = TypeVar("ReturnType")
FunctionType = Callable[..., ReturnType]


class CacheItem:
    __slots__ = ("timestamp", "value")

    def __init__(self, timestamp: int, value: Any):
        self.timestamp = timestamp
        self.value = value


all_cache_pool: Dict[str, Dict] = {}


class MemoryCache(BaseCache):
    cache_pool: Dict[str, CacheItem]

    def __init__(self, *, cached_function: FunctionType, **kwargs):
        super().__init__(cached_function=cached_function, **kwargs)
        if self.limit == -1:
            self.cache_pool = dict()
        else:
            self.cache_pool = LRUDict(self.limit)

    def get(self, *args, **kwargs) -> ReturnType:
        keyword_args, kwargs, cache_key = self.make_key(args, kwargs)
        with self.cache_context(cache_key):
            timestamp = self.get_timestamp()
            cache_info = self.cache_pool.get(cache_key)
            if cache_info and timestamp <= cache_info.timestamp:
                return cache_info.value
            else:
                with self.miss_context(cache_key):
                    value = self.cached_function(*args, **keyword_args, **kwargs)
                    self.set(cache_key, value)
                    return value

    def get_timestamp(self) -> int:
        return int(time.time())

    def set(self, key: str, value: Any) -> None:
        all_cache_pool[self.namespace] = self.cache_pool
        self.cache_pool[key] = CacheItem(
            timestamp=int(time.time()) + self.expire, value=value
        )

    def cache_clear(
        self, args: Optional[tuple] = None, kwargs: Optional[dict] = None
    ) -> int:
        if args or kwargs:
            pattern = self.make_key_pattern(args=args, kwargs=kwargs,)
            count = 0
            for key in filter(pattern.match, list(self.cache_pool.keys())):
                del self.cache_pool[key]
                count += 1
        else:
            count = len(self.cache_pool)
            self.cache_pool.clear()
        return count

    @classmethod
    def get_all_namespace(cls) -> Set[str]:
        return set(all_cache_pool.keys())

    @classmethod
    def flush_cache(cls) -> int:
        count = 0
        for cache_pool in all_cache_pool.values():
            count += len(cache_pool)
            cache_pool.clear()
        return count

    def __del__(self):
        all_cache_pool.pop(self.namespace, None)


class DistributedMemoryCache(DistributedCache):
    cache_pool: Dict[str, CacheItem]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.limit == -1:
            self.cache_pool = dict()
        else:
            self.cache_pool = LRUDict(self.limit)

    def get(self, *args, **kwargs) -> ReturnType:
        keyword_args, kwargs, cache_key = self.make_key(args, kwargs)
        with self.cache_context(cache_key):
            distributed_cache_timestamp: Optional[str] = self.client.get(cache_key)  # type: ignore
            cache_info = self.cache_pool.get(cache_key)
            if distributed_cache_timestamp is None:
                # (first call in first process) or (cache expire)
                with self.miss_context(cache_key):
                    value = self.cached_function(*args, **keyword_args, **kwargs)
                    item = CacheItem(value=value, timestamp=int(time.time()))
                    self.cache_pool[cache_key] = item
                    self.set(cache_key, item)
                return value
            else:
                cache_timestamp = int(distributed_cache_timestamp)
                if cache_info is None:
                    # first call in other processes
                    value = self.cached_function(*args, **keyword_args, **kwargs)
                    self.cache_pool[cache_key] = CacheItem(
                        value=value, timestamp=cache_timestamp
                    )
                    return value
                elif cache_info.timestamp != cache_timestamp:
                    # expire by other process reset cache timestamp
                    with self.miss_context(cache_key):
                        value = self.cached_function(*args, **keyword_args, **kwargs)
                        cache_info.value = value
                        cache_info.timestamp = cache_timestamp
                        return cache_info.value
                else:
                    return cache_info.value

    def set(self, key: str, value: CacheItem) -> None:
        super().set(key, value.timestamp)
        self.cache_pool[key] = value
        all_cache_pool[self.namespace] = self.cache_pool



import unittest
from unittest.mock import Mock

from cache_alchemy import pickle_cache, method_pickle_cache, property_pickle_cache
from tests import CacheTestCase


class TestObject:
    def __init__(self, name: str):
        self.name = name

    def __eq__(self, other):
        return self.name == other.name


return_object = TestObject(name="world")


class PickleCacheTestCase(CacheTestCase):
    def test_cache_function(self):
        call_mock = Mock()

        @pickle_cache
        def hello(name: str) -> TestObject:
            call_mock()
            return return_object

        self.assertEqual(0, hello.cache_clear())
        self.assertEqual(hello("test"), return_object)
        self.assertEqual(1, call_mock.call_count)
        self.assertEqual(1, hello.cache.misses)
        self.assertEqual(0, hello.cache.hits)
        self.assertEqual(hello(name="test"), return_object)
        self.assertEqual(2, hello.cache.misses)
        self.assertEqual(0, hello.cache.hits)
        self.assertEqual(2, call_mock.call_count)
        self.assertEqual(hello(name="test"), return_object)
        self.assertEqual(2, call_mock.call_count)
        self.assertEqual(2, hello.cache.misses)
        self.assertEqual(1, hello.cache.hits)
        self.assertEqual(2, hello.cache_clear())

    def test_cache_method(self):
        call_mock = Mock()

        class Tmp:
            @method_pickle_cache
            def hello(self, name: str) -> TestObject:
                call_mock()
                return return_object

        self.assertEqual(Tmp().hello("test"), return_object)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp().hello("test"), return_object)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp().hello("world"), return_object)
        self.assertEqual(call_mock.call_count, 2)

    def test_cache_class_method(self):
        call_mock = Mock()

        class Tmp:
            @classmethod
            @method_pickle_cache
            def hello(self, name: str) -> TestObject:
                call_mock()
                return return_object

        self.assertEqual(Tmp.hello("test"), return_object)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.hello("test"), return_object)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp.hello("world"), return_object)
        self.assertEqual(call_mock.call_count, 2)

    def test_cache_property(self):
        call_mock = Mock()

        class Tmp:
            @property
            @property_pickle_cache
            def name(self) -> TestObject:
                call_mock()
                return return_object

        self.assertEqual(Tmp().name, return_object)
        self.assertEqual(call_mock.call_count, 1)
        self.assertEqual(Tmp().name, return_object)
        self.assertEqual(call_mock.call_count, 1)


if __name__ == "__main__":
    unittest.main()



from typing import TYPE_CHECKING
from weakref import ref

from configalchemy import BaseConfig

if TYPE_CHECKING:  # pragma: no cover
    from redis import Redis

_current_config_ref = ref(object)


class DefaultConfig(BaseConfig):
    #: default redis url
    CACHE_ALCHEMY_REDIS_URL = "redis://127.0.0.1:6379/0"
    #: distributed json cache backend - default: distributed cache which need assign client to config
    CACHE_ALCHEMY_JSON_BACKEND = "cache_alchemy.backends.json.DistributedJsonCache"
    #: memory cache backend - default: distributed cache which need assign client to config
    CACHE_ALCHEMY_PICKLE_BACKEND = (
        "cache_alchemy.backends.pickle.DistributedPickleCache"
    )
    #: memory cache backend - default: distributed cache which need assign client to config
    CACHE_ALCHEMY_MEMORY_BACKEND = (
        "cache_alchemy.backends.memory.DistributedMemoryCache"
    )
    #: default cache limit per function
    #: - setting to -1 means unlimited
    #: - setting to 0 means uncached
    CACHE_ALCHEMY_DEFAULT_LIMIT = 1000
    #: default cache expire time (seconds)
    #: - setting to 0 means uncached
    CACHE_ALCHEMY_DEFAULT_EXPIRE = 60 * 60 * 24

    #: Need to be assigned after init, if use distributed cache
    cache_redis_client: "Redis"

    def __init__(self):
        super().__init__()
        global _current_config_ref
        _current_config_ref = ref(self)

    @classmethod
    def get_current_config(cls) -> "DefaultConfig":
        current_config = _current_config_ref()
        if not isinstance(current_config, cls):
            raise RuntimeError(f"There is no instance of type {DefaultConfig}")
        return current_config



from typing import Dict, Hashable, List, Set, Optional
from weakref import WeakSet

from .backends.base import BaseCache, CacheFunctionType


class CacheDependency:
    all_dependencies: Dict[Hashable, List["CacheDependency"]] = {}

    def __init__(self, ident: Hashable):
        self.ident = ident
        self.cache_objects: Set[BaseCache] = WeakSet()  # type: ignore

    @classmethod
    def register_dependency(cls, dependency: "CacheDependency") -> None:
        cls.all_dependencies.setdefault(dependency.ident, []).append(dependency)

    @classmethod
    def find_dependencies(cls, ident: Hashable) -> List["CacheDependency"]:
        return cls.all_dependencies.get(ident, [])

    @classmethod
    def dependent_cache_clear(
        cls,
        ident: Hashable,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> int:
        count = 0
        for dependency in cls.find_dependencies(ident):
            for cache in dependency.cache_objects:
                count += cache.cache_clear(args, kwargs)
        return count


class FunctionCacheDependency(CacheDependency):
    """
    Examples::

        @json_cache()
        def add(a, b):
            return a + b

        dependency = FunctionCacheDependency(add)

        @json_cache(dependency=[dependency])
        def add_and_double(a, b):
            return add(a, b) * 2
    """

    def __init__(self, cached_func: CacheFunctionType):
        super().__init__(ident=getattr(cached_func, "cache"))



from types import FunctionType
from typing import Dict, Tuple

# SPECIAL_CHARS
# closing ')', '}' and ']'
# '-' (a range in character set)
# '&', '~', (extended character set operations)
# '#' (comment) and WHITESPACE (ignored) in verbose mode
_special_chars_map = {i: "\\" + chr(i) for i in b"()[]{}?*+-|^$\\.&~# \t\n\r\v\f"}


def escape(pattern):
    """
    Escape special characters in a string.
    """
    if isinstance(pattern, str):
        return pattern.translate(_special_chars_map)
    else:
        pattern = str(pattern, "latin1")
        return pattern.translate(_special_chars_map).encode("latin1")


class UnsupportedError(ValueError):
    pass


def generate_strict_key(
    *, args: Tuple, kwargs: Dict, func: FunctionType, is_method: bool = False
) -> Tuple[dict, dict, str]:
    """Generate function's arguments hash key from optionally typed positional and keyword arguments
    """

    #: The code object representing the compiled function body.
    func_code = getattr(func, "__wrapped__", func).__code__
    #: co_argcount is the number of positional arguments (including arguments with default values)
    #: etc: 2(a and b) in lambda a, b=1, *c, d, e=2, **kwd:...
    pos_count = func_code.co_argcount
    #: co_varnames is a tuple containing the names of the local variables (starting with the argument names)
    #: etc: ('a', 'b', 'd', 'e', 'c', 'kwd') in lambda a, b=1, *c, d, e=2, **kwd:...
    arg_names = func_code.co_varnames
    #: A tuple containing positional arguments names
    #: etc: ('a', 'b') in lambda a, b=1, *c, d, e=2, **kwd:...
    positional = tuple(arg_names[:pos_count])
    #: Number of keyword only arguments (not including ** arg)
    #: etc: 2 in lambda a, b=1, *c, d, e=2, **kwd:...
    keyword_only_count = func_code.co_kwonlyargcount
    #: A tuple containing keyword only arguments names
    #: etc: ('d', 'e') in lambda a, b=1, *c, d, e=2, **kwd:...
    keyword_only = arg_names[pos_count : (pos_count + keyword_only_count)]
    #: A tuple containing default argument values for those arguments that have defaults
    #: etc: (1,) in lambda a, b=1, *c, d, e=2, **kwd:...
    defaults = func.__defaults__ or tuple()
    pos_default_count = len(defaults)
    #: A dict containing defaults for keyword-only parameters.
    #: etc: {'e': 2} in lambda a, b=1, *c, d, e=2, **kwd:...
    kwdefaults = func.__kwdefaults__

    keyword_args = {}
    key = ""

    start = 1 if is_method else 0
    args = args[start:]
    positional = positional[start:]
    # Non-keyword-only parameters w/o defaults.
    non_default_count = pos_count - pos_default_count - start
    while args and non_default_count:
        key += positional[0] + repr(args[0])
        args = args[1:]
        non_default_count -= 1
        positional = positional[1:]

    while args and positional:
        key += positional[0] + repr(args[0])
        args = args[1:]
        defaults = defaults[1:]
        positional = positional[1:]

    while non_default_count:
        value = kwargs.pop(positional[0])
        key += positional[0] + repr(value)
        keyword_args[positional[0]] = value
        positional = positional[1:]
        non_default_count -= 1

    while defaults:
        value = kwargs.pop(positional[0], defaults[0])
        keyword_args[positional[0]] = value
        key += positional[0] + repr(value)
        defaults = defaults[1:]
        positional = positional[1:]

    # *args
    if func_code.co_flags & 4:
        arg_index = pos_count + keyword_only_count
        name = arg_names[arg_index]
        for sub_index, value in enumerate(args[arg_index:]):
            key += f"{name}{sub_index}" + repr(value)

    # Keyword-only parameters.
    for name in keyword_only:
        value = kwargs.pop(name, kwdefaults.get(name))
        keyword_args[name] = value
        key += name + repr(value)

    # **kwargs
    if func_code.co_flags & 8:
        for name, value in sorted(kwargs.items()):
            key += name + repr(value)
    return keyword_args, kwargs, key


def generate_strict_key_pattern(
    *, args: Tuple, kwargs: Dict, func: FunctionType, is_method: bool = False,
) -> str:
    func_code = getattr(func, "__wrapped__", func).__code__
    pos_count = func_code.co_argcount
    arg_names = func_code.co_varnames
    positional = tuple(arg_names[:pos_count])
    keyword_only_count = func_code.co_kwonlyargcount
    keyword_only = arg_names[pos_count : (pos_count + keyword_only_count)]
    fillvalue = r".*?"
    key = ""

    start = 1 if is_method else 0
    args = args[start:]
    positional = positional[start:]

    while positional:
        name = positional[0]
        if args:
            key += name + escape(repr(args[0]))
            args = args[1:]
        elif positional[0] in kwargs:
            key += name + escape(repr(kwargs.pop(name)))
        else:
            key += name + fillvalue
        positional = positional[1:]

    # *args
    if func_code.co_flags & 4:
        arg_index = pos_count + keyword_only_count
        name = arg_names[arg_index]
        for sub_index, value in enumerate(args[arg_index:]):
            key += f"{name}{sub_index}" + escape(repr(value))
        key += fillvalue

    # Keyword-only parameters.
    for name in keyword_only:
        if name in kwargs:
            key += name + escape(repr(kwargs.pop(name)))
        else:
            key += name + fillvalue

    # **kwargs
    if func_code.co_flags & 8:
        for name, value in sorted(kwargs.items()):
            key += name + escape(repr(value))
        key += fillvalue
    return f"{key}$"


def generate_fast_key(
    *, args: Tuple, kwargs: Dict, func: FunctionType, is_method: bool = False
) -> Tuple[dict, dict, str]:
    key = args
    if is_method:
        key = args[1:]
    if kwargs:
        for item in kwargs.items():
            key += item
    return kwargs, {}, repr(key)


def generate_fast_key_pattern(
    *, args: Tuple, kwargs: Dict, func: FunctionType, is_method: bool = False,
) -> str:
    raise UnsupportedError("fast hash not support pattern delete")



import re
from abc import ABC, abstractmethod
from types import FunctionType
from typing import (
    Any,
    ContextManager,
    Dict,
    Tuple,
    cast,
    Pattern,
)
from typing import Callable, TypeVar, Optional, Set, Generic

from ..config import DefaultConfig
from ..utils import (
    generate_strict_key,
    generate_fast_key,
    generate_strict_key_pattern,
    generate_fast_key_pattern,
)

ReturnType = TypeVar("ReturnType")
CacheFunctionType = Callable[..., ReturnType]


class BaseCache(Generic[ReturnType], ABC):
    def __init__(
        self,
        *,
        cached_function: CacheFunctionType,
        expire: int,
        limit: int,
        is_method: bool = False,
        strict=False,
    ):
        self.cached_function = cast(FunctionType, cached_function)
        self.is_method = is_method
        self.expire = expire
        self.limit = limit
        self.hits = self.misses = 0
        self.generate_key = generate_strict_key if strict else generate_fast_key
        self.generate_key_pattern = (
            generate_strict_key_pattern if strict else generate_fast_key_pattern
        )

    @property
    def function_hash(self) -> str:
        return f"{self.__class__.__name__}:{self.cached_function.__module__}:{self.cached_function.__qualname__}"

    @property
    def namespace(self) -> str:
        return f"{self.__class__.__module__}:{self.function_hash}-keys"

    @classmethod
    def get_backend_namespace(cls) -> str:
        return f"{cls.__module__}:{cls.__name__}:all-keys"

    @abstractmethod
    def get(self, *args, **kwargs) -> ReturnType:  # pragma: no cover
        ...

    @abstractmethod
    def set(self, key: str, value: Any) -> None:  # pragma: no cover
        ...

    @abstractmethod
    def cache_clear(
        self, args: Optional[tuple] = None, kwargs: Optional[dict] = None
    ) -> int:  # pragma: no cover
        ...

    @classmethod
    @abstractmethod
    def get_all_namespace(cls) -> Set[str]:  # pragma: no cover
        ...

    @classmethod
    @abstractmethod
    def flush_cache(cls) -> int:  # pragma: no cover
        ...

    def cache_context(self, key: str) -> ContextManager:
        self.hits += 1
        return self

    def miss_context(self, key: str) -> ContextManager:
        self.misses += 1
        self.hits -= 1
        return self

    def make_key(self, args: tuple, kwargs: Dict[str, Any]) -> Tuple[dict, dict, str]:
        keyword_args, kwargs, key = self.generate_key(
            args=args,
            kwargs=kwargs,
            func=self.cached_function,
            is_method=self.is_method,
        )
        return keyword_args, kwargs, f"{self.function_hash}:{key}"

    def make_key_pattern(
        self, args: Optional[tuple], kwargs: Optional[Dict[str, Any]]
    ) -> Pattern:
        pattern = self.generate_key_pattern(
            args=args or tuple(),
            kwargs=kwargs or {},
            func=self.cached_function,
            is_method=self.is_method,
        )
        return re.compile(f"{re.escape(self.function_hash)}:{pattern}", re.DOTALL)

    def __call__(self, *args, **kwargs):
        return self.get(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


DistributedCacheReturnType = TypeVar("DistributedCacheReturnType", bound=bytes)


class DistributedCache(BaseCache[DistributedCacheReturnType]):
    def __init__(self, *, cached_function: FunctionType, **kwargs):
        super().__init__(cached_function=cached_function, **kwargs)
        self.client = DefaultConfig.get_current_config().cache_redis_client
        if self.client.connection_pool.connection_kwargs.get("decode_responses"):
            raise ValueError(
                "Distributed cache client cannot decode response, set decode_responses to False"
            )

    def get(self, *args, **kwargs) -> DistributedCacheReturnType:
        keyword_args, kwargs, cache_key = self.make_key(args, kwargs)
        with self.cache_context(cache_key):
            result = self.client.get(cache_key)
            if result is None:
                with self.miss_context(cache_key):
                    value = self.cached_function(*args, **keyword_args, **kwargs)
                    self.set(cache_key, value)
                    return value
            else:
                return self.deserialize(result)  # type: ignore

    def set(self, key: str, value: DistributedCacheReturnType) -> None:
        self.client.sadd(self.get_backend_namespace(), self.namespace)
        value = self.serialize(value)
        while self.client.scard(self.namespace) >= self.limit:
            del_key = self.client.spop(self.namespace)
            self.client.delete(del_key)

        with self.client.pipeline() as pipe:
            if self.expire == -1:
                pipe.set(key, value)
            else:
                pipe.setex(key, self.expire, value)

            pipe.sadd(self.namespace, key)
            pipe.execute()

    def cache_clear(
        self, args: Optional[tuple] = None, kwargs: Optional[dict] = None
    ) -> int:
        if args or kwargs:
            pattern = self.make_key_pattern(args=args, kwargs=kwargs)
            delete_keys = list(
                filter(
                    pattern.match,
                    map(bytes.decode, self.client.smembers(self.namespace)),
                )
            )
            if delete_keys:
                self.client.delete(*delete_keys)
        else:
            with self.client.pipeline() as pipe:
                delete_keys = self.client.smembers(self.namespace)
                if delete_keys:
                    pipe.delete(*delete_keys)
                    pipe.srem(self.namespace, *delete_keys)
                pipe.srem(self.get_backend_namespace(), self.namespace)
                pipe.execute()
        return len(delete_keys)

    @classmethod
    def get_all_namespace(cls) -> Set[str]:
        client = DefaultConfig.get_current_config().cache_redis_client
        return client.smembers(cls.get_backend_namespace())

    @classmethod
    def flush_cache(cls) -> int:
        client = DefaultConfig.get_current_config().cache_redis_client
        count = 0
        with client.pipeline() as pipe:
            for namespace in cls.get_all_namespace():
                delete_keys = client.smembers(namespace)
                if delete_keys:
                    pipe.delete(*delete_keys)
                    count += len(delete_keys)
                pipe.delete(namespace)
            else:
                pipe.execute()
        return count

    def serialize(
        self, value: DistributedCacheReturnType
    ) -> DistributedCacheReturnType:
        return value

    def deserialize(
        self, result: DistributedCacheReturnType
    ) -> DistributedCacheReturnType:
        return result



import json

from .base import DistributedCache, BaseCache, ReturnType


class DistributedJsonCache(DistributedCache, BaseCache[ReturnType]):
    def serialize(self, value: ReturnType) -> bytes:
        return json.dumps(value).encode()

    def deserialize(self, result: bytes) -> ReturnType:
        return json.loads(result.decode())
